"""
      _                  __
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .core.parsers.env

Provides access to environment variables.

To support nested configs on env vars, chonf provides
a special syntax, consider a python dict that might
be accessed this way:

myconfs['section2']['option8']

A nested structure like this is not supported by env
vars. With chonf you can define this configuration
option by joining the sequence of keys with "__",
adding a prefix to identify the program and isolate
it from other programs:

python:  >>> os.environ['myprogram__section2__option8'] = "value"

posix:   $ myprogram_section2__option8 = "value"

fish:    > set -x myprogram_section2__subsect4__option8 "value"

Then, a program can read this configuration by passing
a list of keys as usual with chonf, passing the program 
prefix in the env_prefix argument:

python:  >>> chonf.get(['section2', 'subsect4', 'option8'], 
         ...           env_prefix='myapp')

The chonf.get() function calls the read() config from each parser
module in chonf.core.parsers until it finds something valid,
The environment variable parser is the first, so it can be used
to quickly short-circuit other configuration sources for quick
tests.

Functions:

    read(keys, *args, **kargs) -> str:
        Tries to read an environment variable, return missing if failed.

"""
from chonf.parsers.env.core import read, list_children



import pytest
import json
import toml
from chonf import Option, load

ENV_PREFIX = "mockapp"

MODEL = {
    "name": Option(),
    "nickname": Option(),
    "email": Option(),
    "system": {
        "background": Option("desert"),
        "theme": Option("dark"),
        "id": Option(),
        "web": {
            "homepages": Option(
                {
                    "job": "job.com",
                    "music": "music.com",
                }
            ),
            "key": "123456",
        },
    },
}


@pytest.fixture(scope="session")
def session_patch():
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()
    # thanks to Rich Inman: https://stackoverflow.com/questions/53963822/python-monkeypatch-setattr-with-pytest-fixture-at-module-scope


@pytest.fixture(scope="session")
def mock_configs(tmp_path_factory, session_patch):
    session_patch.setenv("mockapp__name", "Leibniz Hamilton")
    session_patch.setenv("mockapp__system__theme", "light")
    dir_path = tmp_path_factory.mktemp("mockapp")
    toml_data = {
        "email": "laib.dbz@hotmail.com",
        "system": {"id": "lebzlebz"},
    }
    json_data = {
        "nickname": "Laibilton",
        "system": {
            "background": "flowers",
        },
    }
    with open(dir_path / "config.json", "w") as json_file:
        json.dump(json_data, json_file)
    with open(dir_path / "config.toml", "w") as toml_file:
        toml.dump(toml_data, toml_file)
    return load(MODEL, author="me", name="mockapp", env_prefix="mockapp", path=dir_path)


class TestEnv:
    def test_shallow(self, mock_configs):
        assert mock_configs["name"] == "Leibniz Hamilton"

    def test_nested(self, mock_configs):
        assert mock_configs["system"]["theme"] == "light"


class TestJson:
    def test_shallow(self, mock_configs):
        assert mock_configs["nickname"] == "Laibilton"

    def test_nested(self, mock_configs):
        assert mock_configs["system"]["background"] == "flowers"


class TestToml:
    def test_shallow(self, mock_configs):
        assert mock_configs["email"] == "laib.dbz@hotmail.com"

    def test_nested(self, mock_configs):
        assert mock_configs["system"]["id"] == "lebzlebz"



"""
      _                  __
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .parsers.toml

Provides access to toml files.

The toml syntax provides easy support for nested data. This
data is usually represented by a dictionary that can be
accessed as:

python: >>> toml_data['level1']['level2']['level3']

In a toml file:

    [level1.level2]
    level3 = "value"

With chonf, this will be accessed by:

python: >>> chonf.get(['level1', 'level2', 'level3'], *args, **kwargs)

Except when something else is found first that matches
the keys in some other configuration format, which will
short-circuit the search.

Functions:

    read(keys: List[str], path: pathlib.Path, *args, **kargs) -> Any: 
        Tries to read a configuration option from a Toml file. 

"""
from chonf.parsers.toml.core import read, list_children



"""
      _                  __ 
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .core

Provides the core functionality for chonf.

Functions:

    load(model, env_prefix, dir_path): returns a dictionary constructed based
        on the provided model of configurations, with data read from several
        sources.

Classes:

    Option(default): to be used to create model dictionary to be passed on
        to the load function.

"""
from typing import List, Tuple, Any, Union, Optional, Callable
from functools import partial
from dataclasses import dataclass
from pathlib import Path
from os import PathLike

from chonf.parsers import default_parsers
from chonf.exceptions import (
    NotSubtree,
    SkipSource,
    ConfigReadingError,
    ConfigNotFound,
    ConfigLoadingIncomplete,
    InvalidOption,
)
from chonf.paths import default_path


class BaseOption:
    """Base class for chonf options."""


@dataclass
class Option(BaseOption):
    """Dataclass to represent a config value to be read."""

    default: Any = None
    pre_process: Optional[Callable[[Any], Tuple[bool, Any]]] = None


@dataclass
class Required(BaseOption):
    """Class to represent a required config value to be read."""

    pre_process: Optional[Callable[[Any], Tuple[bool, Any]]] = None


@dataclass
class Repeat(BaseOption):
    """Defines a repeatable substructure within the config model
    with arbitrary root keys."""

    submodel: Union[BaseOption, dict]


@dataclass
class Invalid:
    value: Any
    expected: Any


def load(
    model: dict,
    author: str,
    name: str,
    env_prefix: str = None,
    path: Union[PathLike, str, List[Union[PathLike, str]]] = None,
    *args,
    **kargs,
) -> dict:
    """Load configurations from multiple sources based on
    a model defined as a dictionary.
    """
    if path is None:
        path = default_path(author, name)
    if env_prefix is None:
        env_prefix = name
    if not isinstance(path, list):
        path = [path]

    _search_config = partial(
        search_config, env_prefix=env_prefix, path=path, *args, **kargs
    )

    _repeat_subtree = partial(
        repeat_subtree, env_prefix=env_prefix, path=path, *args, **kargs
    )

    unlocated_keys = []
    invalid_keys = []
    expected_subtree_keys = []

    def _recurse(model, parent_keys=[]):
        configs = {}
        for key, value in model.items():
            keys = parent_keys + [key]
            if isinstance(value, dict):
                config = _recurse(value, keys)
            elif isinstance(value, Option):
                try:
                    config = _search_config(keys, pre_process=value.pre_process)
                except ConfigNotFound:
                    config = value.default
                except InvalidOption as inv_opt:
                    config = inv_opt
                    invalid_keys.append(keys)
            elif isinstance(value, Required):
                try:
                    config = _search_config(keys, pre_process=value.pre_process)
                except ConfigNotFound:
                    config = InvalidOption(None, value)
                    unlocated_keys.append(keys)
                except InvalidOption as inv_opt:
                    config = inv_opt
                    invalid_keys.append(keys)
            elif isinstance(value, Repeat):
                try:
                    config = _recurse(_repeat_subtree(keys, value), keys)
                except InvalidOption as inv_opt:
                    config = inv_opt
                    expected_subtree_keys.append(keys)
            else:
                config = value
            configs[key] = config
        return configs

    configs = _recurse(model)

    if (
        len(unlocated_keys) != 0
        or len(invalid_keys) != 0
        or len(expected_subtree_keys) != 0
    ):
        raise ConfigLoadingIncomplete(
            unlocated_keys, invalid_keys, expected_subtree_keys, configs
        )

    return configs


def search_config(
    keys: List[str],
    env_prefix: str = "",
    path: List[Union[PathLike, str]] = "",
    parsers=default_parsers,
    pre_process: Optional[Callable[[Any], Any]] = None,
    *args,
    **kargs,
) -> Any:
    """Search a configuration variable from multiple sources."""

    if pre_process is None:
        pre_process = lambda x: x
    for p in path:
        p = Path(p)
        for parser in parsers:
            try:
                value = pre_process(
                    parser.read(
                        keys=keys, path=p, env_prefix=env_prefix, *args, **kargs
                    )
                )
                return value
            except SkipSource:
                continue
            except InvalidOption:
                raise
            except Exception as err:
                raise ConfigReadingError from err
    else:
        raise ConfigNotFound


def repeat_subtree(
    keys: List[str],
    repeat: Repeat,
    env_prefix: str = "",
    path: List[Union[PathLike, str]] = "",
    parsers=default_parsers,
    *args,
    **kargs,
) -> dict:
    """Compose repetition of configuration submodel."""
    root_keys = set()
    for p in path:
        p = Path(p)
        for parser in parsers:  # try to get all existing child keys
            try:
                root_keys.update(
                    parser.list_children(
                        keys=keys, path=p, env_prefix=env_prefix, *args, **kargs
                    )
                )
            except NotSubtree as not_subtree:
                raise InvalidOption(not_subtree.value, repeat)
            except Exception as err:
                raise ConfigReadingError from err
    return {key: repeat.submodel for key in root_keys}



from chonf import load, Option
import pytest, toml, json

model = {
    "a": Option(),
    "b": {
        "a": Option(),
        "b": {
            "a": Option(),
            "b": Option(),
        },
    },
}


@pytest.fixture(scope="session")
def system_configs(tmp_path_factory):
    system = tmp_path_factory.mktemp("chonf_multiple_path_system")
    with open(system / "config.toml", "w") as config_file:
        toml.dump({"a": "system", "b": {"a": "system"}}, config_file)
    return system


@pytest.fixture(scope="session")
def user_configs(tmp_path_factory):
    user = tmp_path_factory.mktemp("chonf_multiple_path_user")
    with open(user / "config.json", "w") as config_file:
        json.dump({"b": {"a": "user", "b": {"a": "user"}}}, config_file)
    return user


@pytest.fixture(scope="session")
def mock_configs(user_configs, system_configs):
    configs = load(
        model=model,
        author="me",
        name="asdfghjkl",
        env_prefix="asdfghjkl",
        path=[user_configs, system_configs],  # fist should have greater priority
    )
    return configs


def test_first(mock_configs):
    assert mock_configs["b"]["b"]["a"] == "user"


def test_last(mock_configs):
    assert mock_configs["a"] == "system"


def test_shadow(mock_configs):
    assert mock_configs["b"]["a"] == "user"



"""
      _                  __
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .parsers.json.core

Core functionality for chonf's json configuration parser.

Functions:

    read(keys: List[str], dir_path: pathlib.Path, *args, **kargs) -> Any: 
        Tries to read a configuration option from a json file. 

"""
from functools import cache
from typing import List, Any
from pathlib import Path
import json

from chonf.exceptions import FileAccessError, NotSubtree, SkipSource, InvalidOption


def read(keys: List[str], path: Path, *args, **kwargs) -> Any:
    """Tries to read a configuration option from a Json file.

    Args:
        keys (List[str]): Ordered keys to access the option.
        dir_path: (pathlib.Path): path object representing the file

    Returns:
        Any: Whatever data is found in the file with the given keys.
    """
    config = load_configs(path / "config.json")
    try:
        for key in keys:
            config = config[key]
    except KeyError:
        raise SkipSource
    return config


def list_children(keys: List[str], path: Path, *args, **kwargs) -> Any:
    """Tries to list all children of a configuration node on a Json file.

    Args:
        keys (List[str]): Ordered keys to access the node.
        dir_path: (pathlib.Path): path object representing the file

    Returns:
        Set[str]: list of children keys to requested node, might be empty.

    """
    config = load_configs(path / "config.json")
    try:
        for key in keys:
            config = config[key]
    except KeyError:
        return set()
    if isinstance(config, dict):
        return set(config.keys())
    else:
        raise NotSubtree(config)


@cache
def load_configs(path: Path) -> dict:
    """Reads a Json file into a dict.

    This wraps the load() function from the json package,
    and will use it's default behaviour when called with
    only a file object created by open(path).

    Args:
        path (pathlib.Path): Path to the json file.

    Returns:
        dict: Data read from json file.
    """
    try:
        with open(path) as json_file:
            return json.load(json_file)
    except FileNotFoundError:
        return {}
    except Exception as err:
        raise FileAccessError from err



import pytest
import toml

from chonf.exceptions import SkipSource, NotSubtree
from chonf.parsers.toml import read, list_children

DATA = {
    "option1": "value1",
    "section1": {
        "option2": "value2",
        "repeat": {
            "one": 1,
            "two": 2,
            "three": 3,
        },
    },
    "empty": {},
}


@pytest.fixture(scope="module")
def config_dir(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("chonf_test_toml")
    with open(tmp_path / "config.toml", "w") as toml_file:
        toml.dump(DATA, toml_file)
    return tmp_path


@pytest.fixture(scope="module")
def empty_config_dir(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("chonf_test_toml_empty")
    dir_path = tmp_path / "empty"
    dir_path.mkdir()
    return dir_path


def test_shallow(config_dir):
    key = "option1"
    assert read(keys=[key], path=config_dir) == DATA[key]


def test_shallow_missing(config_dir):
    key = "option3"
    with pytest.raises(SkipSource):
        read(keys=[key], path=config_dir)


def test_nested(config_dir):
    k1, k2 = "section1", "option2"
    assert read(keys=[k1, k2], path=config_dir) == DATA[k1][k2]


def test_nested_missing(config_dir):
    k1, k2 = "section1", "option4"
    with pytest.raises(SkipSource):
        read(keys=[k1, k2], path=config_dir)


def test_missing_file(empty_config_dir):
    with pytest.raises(SkipSource):
        read(keys=["option"], path=empty_config_dir)


def test_list_children_success(config_dir):
    """in this case, list_children should be able to list the child keys"""
    assert list_children(keys=["section1", "repeat"], path=config_dir) == {
        "one",
        "two",
        "three",
    }


def test_list_children_nonexistant(config_dir):
    """in this case, list_children receives a nonexistant key,
    and should return an empty set."""
    assert list_children(keys=["nonexistant"], path=config_dir) == set()


def test_list_children_empty(config_dir):
    """in this case, list_children receives a key to an empty
    dict, and should return an empty set."""
    assert list_children(keys=["empty"], path=config_dir) == set()


def test_list_children_fail(config_dir):
    """in this case, list_children receives a key to a value,
    not a subdict, so it should raise a TypeErr"""
    with pytest.raises(NotSubtree) as err:
        list_children(keys=["option1"], path=config_dir)
    assert err.value.value == "value1"



"""
      _                  __ 
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  core.parsers

Modules to parse different kinds of configurations.

"""
from chonf.parsers import env, toml, json

default_parsers = [env, toml, json]



import pytest
import os

from chonf.exceptions import SkipSource, NotSubtree
from chonf.parsers.env import read, list_children

ENV_PREFIX = "mockapp"

SHALLOW_KEYS = ["some_option"]
SHALLOW_KEY = "__".join([ENV_PREFIX, *SHALLOW_KEYS])

NESTED_KEYS = ["a_section", "other_option"]
NESTED_KEY = "__".join([ENV_PREFIX, *NESTED_KEYS])

VALUE = "something_maybe_important"


def test_shallow(monkeypatch):
    monkeypatch.setenv(SHALLOW_KEY, VALUE)
    assert read(keys=SHALLOW_KEYS, env_prefix=ENV_PREFIX) == VALUE


def test_shallow_empty(monkeypatch):
    monkeypatch.delenv(SHALLOW_KEY, raising=False)
    with pytest.raises(SkipSource):
        read(keys=SHALLOW_KEYS, env_prefix=ENV_PREFIX)


def test_nested(monkeypatch):
    monkeypatch.setenv(NESTED_KEY, VALUE)
    assert read(keys=NESTED_KEYS, env_prefix=ENV_PREFIX) == VALUE


def test_nested_empty(monkeypatch):
    monkeypatch.delenv(NESTED_KEY, raising=False)
    with pytest.raises(SkipSource):
        read(keys=NESTED_KEYS, env_prefix=ENV_PREFIX)


def test_list_children_success(monkeypatch):
    """success case, should return set of keys"""
    monkeypatch.setenv("app__repeat__group1__item1", "1.1")
    monkeypatch.setenv("app__repeat__group1__item2", "1.2")
    monkeypatch.setenv("app__repeat__group2__item1", "2.1")
    monkeypatch.setenv("app__repeat__group2__item2", "2.2")
    assert list_children(keys=["repeat"], env_prefix="app") == {"group1", "group2"}


def test_list_children_empty(monkeypatch):
    """empty case, should return empty set"""
    for key in os.environ:
        if key.startswith("app__repeat"):
            monkeypatch.delenv(key, raising=False)
    assert list_children(keys=["repeat"], env_prefix="app") == set()


def test_list_children_fail(monkeypatch):
    """error case, for when the key is for a value and not a sub-dict"""
    monkeypatch.setenv("app__repeat", "test")
    with pytest.raises(NotSubtree) as err:
        list_children(keys=["repeat"], env_prefix="app")
    assert err.value.value == "test"



import pytest
from chonf import load, Option


@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("mockapp__a", "a")
    monkeypatch.setenv("mockapp__b__b", "b")


def test_by_name(mock_env):
    model = {"a": Option(), "b": {"b": Option()}}
    configs = load(model=model, author="me", name="mockapp")
    assert configs == {"a": "a", "b": {"b": "b"}}


def test_by_prefix(mock_env):
    model = {"a": Option(), "b": {"b": Option()}}
    configs = load(model=model, author="me", name="nope", env_prefix="mockapp")
    assert configs == {"a": "a", "b": {"b": "b"}}



"""
      _                  __
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .parsers.env.core

Core functionality for the environment variable parser.

Functions:

    read(keys, *args, **kargs) -> str:
        Tries to read a configuration option from an environment variable.

"""
from typing import List, Set
import os, re

from chonf.exceptions import SkipSource, NotSubtree


def read(keys: List[str], env_prefix: str, *args, **kargs) -> str:
    """Tries to read a configuration option from an environment variable.

    Args:
        keys (List[str]): Sequence of keys to construct the env var name.
        env_prefix (str): Prefix of the env var name

    Returns:
        str: The env var value.

    Raises:
        SkipSource: For expected problems, to be ignored and skipped.

    """
    try:
        return os.environ["__".join([env_prefix, *keys])]
    except KeyError:
        raise SkipSource


def list_children(keys: List[str], env_prefix: str, *args, **kargs) -> Set[str]:
    parent_key = "__".join([env_prefix, *keys])
    children = set()
    for key, value in os.environ.items():
        if key == parent_key:
            raise NotSubtree(value)  # error when key is for a leaf node (not a dict)
        elif key.startswith(parent_key):
            child = re.split("__", key)[len(keys) + 1]
            children.add(child)
    return children



"""
      _                  __ 
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  

Chooses configurations from multiple posible sources.

Functions:
    
    get(key, dir_path, *args, **kargs): Tries to get some user configuration

"""
from chonf.core import load, Option, Required, Repeat
from chonf.exceptions import ConfigReadingError, ConfigLoadingIncomplete, InvalidOption
from chonf.paths import default_path



"""
      _                  __ 
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .core.paths

Provides default config paths for chonf.

Variables:

    default_path(author: str, name: str) -> List[pathlib.Path]:
        returns the paths to the user-wide and system-wide default
        configuration directories for the program running on the
        current system.

"""

from pathlib import Path
from typing import List
from os.path import expandvars
from functools import cache
from chonf.exceptions import UnrecognizedOS
from platform import system


@cache
def default_path(author: str, name: str) -> List[Path]:
    """Default config directory paths.

    Get the paths to the user-wide and system-wide default
    configuration directories for the program running on the
    current system.

    Args:
        author (str): Name of the author (company or person)
            of the program, used by Windows's default for
            program data files.
        name (str): Name of the program itself, that names
            the configuration directory under the system's
            default.

    """
    os = system()
    if os == "Linux":
        return [
            Path("~/.config/").expanduser() / name,
            Path("/etc/") / name,
        ]
    elif os == "Windows":
        return [
            Path(expandvars("%APPDATA%")) / author / name,
            Path(expandvars("%PROGRAMDATA%")) / author / name,
        ]
    elif os == "Darwin":  # macos
        return [
            Path("~/Library/Preferences/").expanduser() / name,
            Path("/Library/Preferences/") / name,
        ]
    else:
        raise UnrecognizedOS



"""
      _                  __ 
  ___| |__   ___  _ __  / _|
 / __| '_ \ / _ \| '_ \| |_ 
| (__| | | | (_) | | | |  _|
 \___|_| |_|\___/|_| |_|_|  .core.exceptions

Provides exceptions for chonf.

Classes (exceptions):

    ConfigReadingError: Raised when an attempt to read a configuration fails
        in an unexpected manner.

    FileAccessError: Raised when an attempt to access a config file fails
        in an unexpected manner.

    SkipSource: Raised when an attempt to access a config source fails in an
        expected manner, so that chonf can continue looking elsewhere.

"""


from dataclasses import dataclass
from typing import List, Any


class ConfigReadingError(Exception):
    """Throw when an attempt to access a configuration option fails
    in an unexpected manner, for a more comprehensive error log.
    """

    pass


class FileAccessError(Exception):
    """Throw when an attempt to read a configuration file fails in an
    unexpected manner, for a more comprehensive error log.

    Should not be confused with FileNotFoundError, as this exception
    is expected and should usually cause a SkipSource exception to
    be raised.
    """

    pass


class SkipSource(Exception):
    """Throw when an attempt to access a configuration source fails
    in an expected way, to signify the getter to continue looking at
    other sources without stopping the program.
    """

    pass


class ConfigNotFound(Exception):
    """Throw when an attempt to find a configuration option fails in
    all sources.
    """


@dataclass
class ConfigLoadingIncomplete(Exception):
    """Throw when an attempt to load configurations fails to
    attend the defined constraints.
    """

    unlocated_keys: List[List[str]]
    invalid_keys: List[List[str]]
    expected_subtree_keys: List[List[str]]
    loaded_configs: dict


class UnrecognizedOS(Exception):
    """Throw when the operational system is not recognized."""


@dataclass
class InvalidOption(Exception):
    """Throw when a Option with validation finds a invalid value.
    Also works as a information dataclass to inform the problem
    in the configurations."""

    value: Any
    expected: Any


@dataclass
class NotSubtree(Exception):
    """Throw when trying to read a subtree structure such as a
    repeat, and finding a leaf node.
    """

    value: Any



from pyparsing import Opt
from chonf import load, Option, Required, ConfigLoadingIncomplete
import pytest
import json
from functools import partial

from chonf.exceptions import InvalidOption


@pytest.fixture(scope="module")
def preset_load(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("chonf_pre_processing_test")
    data = {
        "a": "1",
        "b": {"a": 1, "b": {"a": 1.5, "b": {"a": "one"}}},
        "c": "1.0",
        "d": "1.5",
        "e": 1.0,
    }
    with open(tmp_dir / "config.json", "w") as f:
        json.dump(data, f)
    return partial(
        load,
        author="me",
        name="myapp",
        path=tmp_dir,
    )


def test_numeric_success(preset_load):
    """Tests a simple pre-processing function that tries
    to convert certain options to floats, in this case they should
    succeed.
    """

    def into_float(value):
        try:
            return float(value)
        except ValueError:
            raise InvalidOption(value, "convertible to float") from ValueError

    model = {
        "a": Option(pre_process=into_float),
        "b": {
            "a": Option(pre_process=into_float),
            "b": {"a": Required(pre_process=into_float), "b": {"a": Required()}},
        },
    }
    config = preset_load(model=model)
    assert config == {"a": 1.0, "b": {"a": 1.0, "b": {"a": 1.5, "b": {"a": "one"}}}}


def test_numeric_failure(preset_load):
    """Tests a simple pre-processing function that tries
    to convert ceretain options to integers, in this case some
    should fail and some should succeed.
    """

    def into_int(value):
        # test if number can be read as integer, even if it has a dot
        try:
            fvalue = float(value)
            ivalue = int(fvalue)
            if fvalue == ivalue:
                return ivalue
            else:
                raise InvalidOption(value, "convertible to int")
        except ValueError as err:
            raise InvalidOption(value, "convertible to int") from err

    model = {
        "a": Option(pre_process=into_int),
        "b": {
            "a": Option(pre_process=into_int),
            "b": {
                "a": Required(pre_process=into_int),
                "b": {
                    "a": Required(pre_process=into_int),
                },
            },
        },
        "c": Option(pre_process=into_int),
        "d": Option(pre_process=into_int),
        "e": Option(pre_process=into_int),
    }
    with pytest.raises(ConfigLoadingIncomplete) as err:
        preset_load(model=model)
    assert err.value.loaded_configs == {
        "a": 1,
        "b": {
            "a": 1,
            "b": {
                "a": InvalidOption(1.5, "convertible to int"),
                "b": {
                    "a": InvalidOption("one", "convertible to int"),
                },
            },
        },
        "c": 1,
        "d": InvalidOption("1.5", "convertible to int"),
        "e": 1,
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chonf',
 'chonf.parsers',
 'chonf.parsers.env',
 'chonf.parsers.env.test',
 'chonf.parsers.json',
 'chonf.parsers.json.test',
 'chonf.parsers.toml',
 'chonf.parsers.toml.test',
 'chonf.test']

package_data = \
{'': ['*']}

install_requires = \
['toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'chonf',
    'version': '0.1.1',
    'description': 'User config management made simple and powerful',
    'long_description': '```\n      _                  __ \n  ___| |__   ___  _ __  / _|\n / __| \'_ \\ / _ \\| \'_ \\| |_ \n| (__| | | | (_) | | | |  _|\n \\___|_| |_|\\___/|_| |_|_|\n```\n\nAn early concept for a opinionated configuration framework to give\npower users the freedom to use whatever config file syntax they like.\nIt\'s written in python.\n\n**Chonf** enforces a structure of configuration that can always be\ntranslated into a tree, such as nested dictionaries in python. \nMost of the widely used config file syntaxes work like this: Json, \nToml, Yaml, and so on.\n\n# How To\n\nThe configuration loading in your program should look like this:\n\n```python\nfrom chonf import load, Option\n\n# Create a configuration model as a dictionary\nmymodel = {\n    "ui_theme": Option("dark"), # default value here is "dark"\n    "user_info": {\n        "name": Option(), # empty Option, default here is None\n        "email": Option(),\n    }\n}\n\nmyconfigs = load(\n    mymodel, \n    author="mycompany",\n    name="myprogram",\n)\n\n# myconfigs is now a dict of your configurations\n```\n\nA Toml config file, for example, could look like this:\n```toml\nui_theme = "light"\n\n[user_info]\nname = "Tom Preston-Werner"\nemail = "tomate@hotmail.com"\n```\n\n## Overwriting stuff with environment variables\n\nFor **Chonf**, environment variables have higher *precedence*\nthan config files such as Toml or Json files, so you can quickly\noverwrite a configuration set in a config file with a environment\nvariable such as:\n\n```console\nfoo@bar:~$ export myprogram__ui_theme="adwaita"\n```\n\nThis allows for quick tests in config changes without opening, editing\nand saving config files multiple times.\n\nThe syntax for the env variable names is the following: first the \nspecified `env_prefix` argument on the load function, than the keys\nfrom the root of the dictionary tree down to the name of the option itself,\nseparated by ***double__underscores***.\n\nOn the previous example, this is how you would name the environment\nvariables:\n\n- `myconfigs["ui_theme"]` is named `myprogram__ui_theme`\n- `myconfigs["user_info"]["name"]` is named `myprogram__user_info__name`\n- `myconfigs["user_info"]["email"]` is named `myprogram__user_info__email`\n\nThis unusual double underscore syntax allows usage of single underscores\nas word separators on the option names without creating ambiguity. \n\nNote that the default for environment variables is to use only letters,\ndigits and underscores, so it\'s advisable to use only these characters\nfor naming the model dictionary keys. Otherwise, users might not be able\nto easily access those options through the shell.\n\n## Required options\n\nFrom **Chonf** you can also import a `Required` dataclass that will\nnot take a default value, and will cause the `load()` function to\nraise a `ConfigLoadingIncomplete` exception if the option is not found.\nThis exception is also a\ndataclass that will contain the data that was read, with `Required`\nobjects left where stuff was not found, a list of all the keys for\nthe unlocated options and also invalid options \n(see *Functional Pre-Processing and Validation*). \nAs an example, if your code looks like this:\n\n```python\nfrom chonf import load, Option, Required\n\nmodel = {\n    "a": Required(),\n    "b": Option(),\n    "c": {\n        "c_a": Required(),\n        "c_b": Option(),\n    }\n}\n\nconf = load(model, ... # your program arguments would be here\n\n```\n\n, if the option `conf["a"]` or `conf["c"]["a"]` are missing,\nthe `load()` function will raise a `ConfigLoadingIncomplete` \nexception. In case all options are missing (following comments \nrepresent output):\n\n```python\ntry:\n    conf = load(model, ... # your program arguments would be here\nexcept ConfigLoadingIncomplete as err:\n    print(err.unlocated_keys) \n    # [["a"], ["c","c_a"]]\n\n    print(err.loaded_configs) \n    # {\n    #   "a": InvalidOption(value=None, expected=Required()),\n    #   "b": None,\n    #   "c": {\n    #       "c_a": InvalidOption(value=None, expected=Required()),\n    #       "c_b": None\n    #   }\n    # }\n\n```\n\n# Multiple Config Paths\n\nIf you decide to offer more than one option of config file location,\npass a list of paths instead:\n\n```python\nconfigs = load(\n    model = mymodel,\n    author = "mycompany",\n    name = "myprogram",\n    path = ["/home/me/.config/myprogram", "/etc/myprogram"],\n)\n```\n\nYou can have several config directory options. What comes first in the\nlist will have higher priority. In this example, the user level\nconfigurations will be able to shadow the ones defined system-wide.\n\n# Change Environment Variables Prefix\n\nIf you would like to use something other than the name of your\nprogram as the prefix for the env vars, pass the `env_prefix`\nargument, as the name is also used in the path to the default\nconfig directories.\n\n```python\nconfigs = load(\n    model = mymodel,\n    author = "mycompany",\n    name = "myprogram",\n    env_prefix = "mypro",\n    ...\n```\n\n# Functional Pre-Processing and Validation\n\nIf some of your options require some specific type of data or\nany other kind of validation, **Chonf** provides a functional\napproach to verifying and pre-processing your options during\nloading.\n\nPass to your option a callable object (like a function) as the\n`pre_process` argument. The function should be able to receive\na single positional argument (the value read from some env var\nor file) and returning the pre-processed value. If the value\nis invalid, the function should raise a `chonf.InvalidOption`\nexception containing the value received and some info about\nwhat was expected.\n\nIn the following snippet, we can check if a option is a number\nand immediately convert it into it\'s numeric type:\n\n```python\nfrom chonf import load, Option, InvalidOption\n\ndef into_number(value):\n    try:\n        return float(value)\n    except ValueError as err:\n        raise InvalidOption(value, "something compatible with float") from err\n\nmodel = {"username": Option(), "a_number": Option(pre_process=into_number)}\n```\n\nFuture versions of **Chonf** will implement common predefined pre-process\nfunctions.\n\n# Repeating Structures\n\nSometimes you might want to have a group of many similar configuration\nstructures somewhere in your model. For example, in a text editor,\nyou might have a group of language server definitions, with each one\'s\nspecific info. A simple example:\n\n```toml\n[language_servers]\n\n    [language_servers.python]\n    name = "Python Language Server"\n    command = "pyls"\n\n    [language_servers.fortran]\n    name = "Fortran Language Server"\n    command = "flang"\n```\n\n**Chonf** provides a simple way to define such repetitions. Define\nthe repeating sub-model in a `Repeat` object:\n\n```python\nfrom chonf import load, Required, Repeat\n\nmodel = {\n    "language_servers": Repeat({\n        "name": Required(),\n        "command": Required()\n    })\n}\n\nconfigs = load(model, "mycompany", "myapp")\n```\n\nNotice how you can have required options inside the repeating structures.\nThe blocks for python and fortran in the previous example are not required,\nbut if they are found, their name and command will be required. Also, \n`Repeat` must find a subtree on your configurations, if it finds a leaf node\n(a value such as a string) it will deem it invalid.\n\nAlso, if you know the keys for each block, nothing is stopping you from\nusing dictionary comprehensions.\n\n# Default Config Paths\n\n**Chonf** should be able to identify the user\'s OS and locate\ndefault locations for user-wide and system-wide configurations, \nbut this feature is not well tested yet. Feedback is very welcome. \n\nTo try it, just skip the `path` argument on the `load()`\nfunction, and it will try to find the config files in your system\'s\ndefault. If you wish to see where this location might be\nby **Chonf**\'s algorithm, call `chonf.default_path()` or\ntake a look at the `paths.py` module on the source code.\n\n# Next Steps\n\nThis project is still in its early stages, this is what\nwe plan to allow next:\n\n- Procedurally generated options from python scripts \n- Yaml and XML (.ini is way too limiting so we probably won\'t implement it)\n- User and System level priority order customization for the end user\n- Support for custom relative paths (location of config files inside config folder)\n- Function for dumping data back into whatever format the end user defined as preferred\n- Make all file formats optional other than Json that already comes with python. \n- Figure out versioning and publish on pip\n',
    'author': 'João C. Rodrigues Jr.',
    'author_email': 'jc.rodrigues1997@usp.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/Ovenbird-j/chonf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from chonf import load, Required, ConfigLoadingIncomplete
import pytest


@pytest.fixture
def env_vars(monkeypatch):
    monkeypatch.setenv("mockapp__aaa", "value_a")
    monkeypatch.setenv("mockapp__bbb__ccc", "value_b")


def test_read(env_vars):
    model = {"aaa": Required(), "bbb": {"ccc": Required()}}
    conf = load(model, author="me", name="mockapp", env_prefix="mockapp")
    assert conf["aaa"] == "value_a" and conf["bbb"]["ccc"] == "value_b"


def test_fail(env_vars):
    model = {"aaa": Required(), "bbb": {"ccd": Required()}, "ccc": Required()}
    with pytest.raises(ConfigLoadingIncomplete) as err:
        load(model, author="me", name="mockapp", env_prefix="mockapp")
    assert err.value.unlocated_keys == [["bbb", "ccd"], ["ccc"]]



from chonf import load, Option, Required, Repeat, ConfigLoadingIncomplete
import pytest
import json

from chonf.exceptions import InvalidOption

MODEL = {
    "option_a": Option(),
    "option_b": Option(),
    "option_repeat": Repeat(Option()),
    "users": Repeat(
        {
            "name": Required(),
            "nick": Option(),
        }
    ),
}

VALID_DATA = {
    "option_a": "value_a",
    "option_b": "value_b",
    "option_repeat": {"a": "a", "b": "b", "c": "c"},
    "users": {
        "user_1": {
            "name": "Jean Jacques Rousseau",
            "nick": "Jack",
        },
        "user_2": {
            "name": "Mary Wollstonecraft",
        },
        "user_3": {"name": "James Hoffman", "nick": "Internet Coffee Dude"},
    },
}

INVALID_DATA = {
    "option_a": "value_a",
    "option_b": "value_b",
    "option_repeat": "invalid_value",  # this should be a subtree
    "users": {
        "user_1": {
            "name": "Jean Jacques Rousseau",
            "nick": "Jack",
        },
        "user_2": {
            "nick": "Did she have a nickname?",  # no name on this one
        },
        "user_3": {"name": "James Hoffman", "nick": "Internet Coffee Dude"},
    },
}


@pytest.fixture(scope="module")
def valid_conf_dir(tmp_path_factory):
    dir_path = tmp_path_factory.mktemp("chonf_repeat_test_valid")
    with open(dir_path / "config.json", "w") as f:
        json.dump(VALID_DATA, f)
    return dir_path


@pytest.fixture(scope="module")
def invalid_conf_dir(tmp_path_factory):
    dir_path = tmp_path_factory.mktemp("chonf_repeat_test_invalid")
    with open(dir_path / "config.json", "w") as f:
        json.dump(INVALID_DATA, f)
    return dir_path


def test_valid_data(valid_conf_dir):
    configs = load(MODEL, author="me", name="mock", path=valid_conf_dir)
    assert configs == {
        "option_a": "value_a",
        "option_b": "value_b",
        "option_repeat": {"a": "a", "b": "b", "c": "c"},
        "users": {
            "user_1": {
                "name": "Jean Jacques Rousseau",
                "nick": "Jack",
            },
            "user_2": {
                "name": "Mary Wollstonecraft",
                "nick": None,  # here the not found value reads to None
            },
            "user_3": {"name": "James Hoffman", "nick": "Internet Coffee Dude"},
        },
    }


@pytest.fixture(scope="module")
def invalid_data_error(invalid_conf_dir):
    with pytest.raises(ConfigLoadingIncomplete) as err:
        load(MODEL, author="me", name="mock", path=invalid_conf_dir)
    return err.value


def test_invalid_data_loaded_configs(invalid_data_error):
    assert invalid_data_error.loaded_configs == {
        "option_a": "value_a",
        "option_b": "value_b",
        "option_repeat": InvalidOption("invalid_value", Repeat(Option())),
        "users": {
            "user_1": {
                "name": "Jean Jacques Rousseau",
                "nick": "Jack",
            },
            "user_2": {
                "name": InvalidOption(None, Required()),
                "nick": "Did she have a nickname?",  # no name on this one
            },
            "user_3": {"name": "James Hoffman", "nick": "Internet Coffee Dude"},
        },
    }
